/*
 * Copyright (C) 2001-2006 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#if !defined(CLIENT_H)
#define CLIENT_H

#if _MSC_VER > 1000
#pragma once
#endif

#include "User.h"
#include "BufferedSocket.h"
#include "SettingsManager.h"
#include "TimerManager.h"
#include "DebugManager.h"

class Client;
class AdcCommand;
class ClientManager;
class ClientListener
{
public:
        virtual ~ClientListener() { }
        template<int I> struct X { enum { TYPE = I };  };

        typedef X<0> Connecting;
        typedef X<1> Connected;
        typedef X<3> UserUpdated;
        typedef X<4> UsersUpdated;
        typedef X<5> UserRemoved;
        typedef X<6> Redirect;
        typedef X<7> Failed;
        typedef X<8> GetPassword;
        typedef X<9> HubUpdated;
        typedef X<11> Message;
        //typedef X<12> StatusMessage;
        typedef X<13> PrivateMessage;
        typedef X<14> UserCommand;
        typedef X<15> HubFull;
        typedef X<16> NickTaken;
        typedef X<17> SearchFlood;
        typedef X<18> NmdcSearch;
        typedef X<19> AdcSearch;
        typedef X<20> CheatMessage;
        typedef X<21> HubTopic;

        virtual void on(Connecting, Client*) throw() { }
        virtual void on(Connected, Client*) throw() { }
        virtual void on(UserUpdated, Client*, const User::Ptr&) throw() { } // !SMT!-fix
        virtual void on(UsersUpdated, Client*, const User::PtrList&) throw() { }
        virtual void on(UserRemoved, Client*, const User::Ptr&) throw() { }
        virtual void on(Redirect, Client*, const string&) throw() { }
        virtual void on(Failed, Client*, const string&) throw() { }
        virtual void on(GetPassword, Client*) throw() { }
        virtual void on(HubUpdated, Client*) throw() { }
        virtual void on(Message, Client*, const User::Ptr&, const string&) throw() { }
        //virtual void on(StatusMessage, Client*, const string&) throw() { }
        virtual void on(PrivateMessage, Client*, const User::Ptr, const User::Ptr, const User::Ptr, const string&, bool = true) throw() { } // !SMT!-S
        virtual void on(UserCommand, Client*, int, int, const string&, const string&) throw() { }
        virtual void on(HubFull, Client*) throw() { }
        virtual void on(NickTaken, Client*) throw() { }
        virtual void on(SearchFlood, Client*, const string&) throw() { }
        virtual void on(NmdcSearch, Client*, const string&, int, int64_t, int, const string&, bool) throw() { }
        virtual void on(AdcSearch, Client*, const AdcCommand&, const CID&) throw() { }
        virtual void on(CheatMessage, Client*, const string&) throw() { }
        virtual void on(HubTopic, Client*, const string&) throw() { }
};

/** Yes, this should probably be called a Hub */
class Client : public Speaker<ClientListener>, public BufferedSocketListener, protected TimerManagerListener {
public:
        typedef Client* Ptr;
        typedef slist<Ptr> List;
        typedef List::const_iterator Iter;

        virtual void connect();
        virtual void disconnect(bool graceless);

        virtual void connect(const OnlineUser& user) = 0;
        virtual void hubMessage(const string& aMessage) = 0;
        virtual void privateMessage(const User::Ptr& user, const string& aMessage, bool annoying = true) = 0; // !SMT!-S
        virtual void sendUserCmd(const string& aUserCmd) = 0;
        virtual void search(int aSizeMode, int64_t aSize, int aFileType, const string& aString, const string& aToken) = 0;
        virtual void password(const string& pwd) = 0;
        virtual void info() = 0;
        virtual void cheatMessage(const string& aLine) = 0;

        virtual size_t getUserCount() const = 0;
        int64_t getAvailable() const { return availableBytes; };
        virtual OnlineUser* getUser(const User::Ptr& aUser); // !SMT!-S
        //virtual OnlineUser* findUser(const string& aNick) = 0; // !SMT!-S

        virtual void send(const AdcCommand& command) = 0;

        virtual string escape(string const& str) const { return str; }

        bool isConnected() const { return socket && socket->isConnected(); }
        bool isOp() const { return getMyIdentity().isOp(); }

        virtual void refreshUserList(bool) = 0;

        short getPort() const { return port; }
        const string& getAddress() const { return address; }

        const string& getIp() const { return ip; }
        string getIpPort() const { return getIp() + ':' + Util::toString(port); }
        string getLocalIp() const;

        void updated(const User::Ptr& aUser) { fire(ClientListener::UserUpdated(), this, aUser); } // !SMT!-fix

        static int getTotalCounts() {
                return counts.normal + counts.registered + counts.op;
        }

        static string getCounts() {
                char buf[128];
                return string(buf, snprintf(buf, sizeof(buf), "%ld/%ld/%ld", counts.normal, counts.registered, counts.op));
        }

        string getRawCommand(const int aRawCommand) {
                switch(aRawCommand) {
                        case 1: return rawOne;
                        case 2: return rawTwo;
                        case 3: return rawThree;
                        case 4: return rawFour;
                        case 5: return rawFive;
                }
                return Util::emptyString;
        }
        StringMap& escapeParams(StringMap& sm) {
                for(StringMapIter i = sm.begin(); i != sm.end(); ++i) {
                        i->second = escape(i->second);
                }
                return sm;
        }

        void reconnect();
        void shutdown();
        bool isActive();

        void send(const string& aMessage) { send(aMessage.c_str(), aMessage.length()); }
        void send(const char* aMessage, size_t aLen) {
                //dcassert(socket);
                if(!socket)
                        return;
                updateActivity();
                socket->write(aMessage, aLen);
                COMMAND_DEBUG(aMessage, DebugManager::HUB_OUT, getIpPort());
        }

        string getMyNick() const { return getMyIdentity().getNick(); }
        string getHubName() const { return getHubIdentity().getNick().empty() ? getHubUrl() : getHubIdentity().getNick(); }
        string getHubDescription() const { return getHubIdentity().getDescription(); }

        Identity& getHubIdentity() { return hubIdentity; }

        const string& getHubUrl() const { return hubUrl; }

        GETSET(Identity, myIdentity, MyIdentity);
        GETSET(Identity, hubIdentity, HubIdentity);

        GETSET(string, defpassword, Password);
        GETSET(uint32_t, reconnDelay, ReconnDelay);
        GETSET(uint32_t, lastActivity, LastActivity);
        GETSET(bool, registered, Registered);
        GETSET(bool, autoReconnect, AutoReconnect);

        GETSET(string, currentNick, CurrentNick);
        GETSET(string, currentDescription, CurrentDescription);
        GETSET(string, currentEmail, CurrentEmail);

        GETSET(bool, stealth, Stealth);
        GETSET(string, rawOne, RawOne);
        GETSET(string, rawTwo, RawTwo);
        GETSET(string, rawThree, RawThree);
        GETSET(string, rawFour, RawFour);
        GETSET(string, rawFive, RawFive);
        GETSET(string, favIp, FavIp);
        GETSET(string, clientId, ClientId); // !SMT!-S
        
		// No Hide Share Mod	
		bool hideShare; 	
		bool getHideShare() {
			return hideShare;
		}
		void setHideShare(bool) {
			hideShare = false ;
		}
		int64_t availableBytes;
protected:
        friend class ClientManager;
        Client(const string& hubURL, char separator, bool secure_);
        virtual ~Client() throw();
        struct Counts {
                Counts(long n = 0, long r = 0, long o = 0) : normal(n), registered(r), op(o) { }
                volatile long normal;
                volatile long registered;
                volatile long op;
                bool operator !=(const Counts& rhs) { return normal != rhs.normal || registered != rhs.registered || op != rhs.op; }
        };

        enum States {
                STATE_CONNECTING,       ///< Waiting for socket to connect
                STATE_PROTOCOL,         ///< Protocol setup
                STATE_IDENTIFY,         ///< Nick setup
                STATE_VERIFY,           ///< Checking password
                STATE_NORMAL,           ///< Running
                STATE_DISCONNECTED,     ///< Nothing in particular
        } state;

        BufferedSocket* socket;

        static Counts counts;
        Counts lastCounts;

        void updateCounts(bool aRemove);
        void updateActivity() { lastActivity = GET_TICK(); }

        /** Reload details from favmanager or settings */
        void reloadSettings(bool updateNick);

        virtual string checkNick(const string& nick) = 0;

        // TimerManagerListener
        virtual void on(Second, uint32_t aTick) throw();
        // BufferedSocketListener
        virtual void on(Connecting) throw() { fire(ClientListener::Connecting(), this); }
        virtual void on(Connected) throw();
        virtual void on(Line, const string& aLine) throw();
        virtual void on(Failed, const string&) throw();

private:

        enum CountType {
                COUNT_UNCOUNTED,
                COUNT_NORMAL,
                COUNT_REGISTERED,
                COUNT_OP
        };

        Client(const Client&);
        Client& operator=(const Client&);

        string hubUrl;
        string address;
        string ip;
        uint16_t port;
        char separator;
        bool secure;
        CountType countType;
};

#endif // !defined(CLIENT_H)

/**
 * @file
 * $Id: Client.h,v 1.65 2006/11/04 14:08:28 bigmuscle Exp $
 */
